class ImageStatus:
    New = "NEW"
    Done = "DONE"
    Skipped = "SKIPPED"
    InProgress = "IN PROGRESS"
    ToReview = "TO REVIEW"
    AutoLabelled = "AUTO-LABELLED"


class ExportFormat:
    JSON_v11 = "json_v1.1"
    SEMANTIC_PNG = "semantic_png"
    JSON_COCO = "json_coco"
    IMAGES = "images"


class SemanticFormat:
    GS_DESC = "gs_desc"
    GS_ASC = "gs_asc"
    CLASS_COLOR = "class_color"


class SemanticOrder:
    Z_INDEX = "z_index"
    CLASS_TYPE = "class_type"
    CLASS_ORDER = "class_order"


WAIT_INTERVAL_SEC = 10

VALID_STATUSES = [ImageStatus.New, ImageStatus.Done, ImageStatus.Skipped, ImageStatus.InProgress, ImageStatus.ToReview,
                  ImageStatus.AutoLabelled]
VALID_EXPORT_FORMATS = [ExportFormat.JSON_v11, ExportFormat.SEMANTIC_PNG, ExportFormat.JSON_COCO, ExportFormat.IMAGES]
VALID_SEMANTIC_FORMATS = [SemanticFormat.GS_DESC, SemanticFormat.GS_ASC, SemanticFormat.CLASS_COLOR]
VALID_SEMANTIC_ORDER = [SemanticOrder.Z_INDEX, SemanticOrder.CLASS_TYPE, SemanticOrder.CLASS_ORDER]



from collections import OrderedDict

from .exception import LimitExceededException, ValidationException
from .hasty_object import HastyObject
from .tag_class import TagClass


C_TAGS_LIMIT = 100


class Tag(HastyObject):
    endpoint = '/v1/projects/{project_id}/images/{image_id}/tags'

    def __repr__(self):
        return self.get__repr__(OrderedDict({"id": self._id, "tag_class_id": self._tag_class_id,
                                             "tag_class_name": self._tag_class_name, "image_id": self._image_id}))

    @property
    def id(self):
        """
        :type: string
        """
        return self._id

    @property
    def tag_class_id(self):
        """
        :type: string
        """
        return self._tag_class_id

    @property
    def project_id(self):
        """
        :type: string
        """
        return self._project_id

    @property
    def image_id(self):
        """
        :type: string
        """
        return self._image_id

    @property
    def tag_class_name(self):
        """
        :type: string
        """
        return self._tag_class_name

    def _init_properties(self):
        self._id = None
        self._tag_class_id = None
        self._tag_class_name = None
        self._project_id = None
        self._image_id = None

    def _set_prop_values(self, data):
        if "id" in data:
            self._id = data["id"]
        if "tag_class_name" in data:
            self._tag_class_name = data["tag_class_name"]
        if "tag_class_id" in data:
            self._tag_class_id = data["tag_class_id"]
        if "project_id" in data:
            self._project_id = data["project_id"]
        if "image_id" in data:
            self._image_id = data["image_id"]

    @staticmethod
    def _batch_create(requester, project_id, image_id, tags):
        if len(tags) == 0:
            return []
        data = []
        if len(tags) > C_TAGS_LIMIT:
            raise LimitExceededException.max_tags_per_batch(len(tags))
        for tag in tags:
            if isinstance(tag, TagClass):
                data.append({"tag_class_id": tag.id})
            elif "tag_class_id" not in tag:
                raise ValidationException("tag_class_id property should be defined")
            else:
                data.append({"tag_class_id": tag["tag_class_id"]})
        res = requester.post(Tag.endpoint.format(project_id=project_id, image_id=image_id), json_data=data)
        new_tags = []
        for tag in res:
            new_tags.append(Tag(requester, tag, {"project_id": project_id, "image_id": image_id}))
        return new_tags

    @staticmethod
    def _batch_delete(requester, project_id, image_id, tags):
        data = []
        for tag in tags:
            if isinstance(tag, TagClass):
                data.append({"tag_class_id": tag.id})
            elif isinstance(tag, Tag):
                data.append({"tag_id": tag.id})
            elif "tag_class_id" not in tag and "tag_id" not in tag:
                raise ValidationException("tag_class_id or tag_id property should be defined")
            else:
                data.append({"tag_id": tag.get("tag_id"),
                             "tag_class_id": tag.get("tag_class_id")})
        requester.delete(Tag.endpoint.format(project_id=project_id, image_id=image_id), json_data=data)



from collections import OrderedDict

from .hasty_object import HastyObject


class Workspace(HastyObject):
    """Class that contains some basic requests and features for workspaces."""
    endpoint = '/v1/workspaces'
    endpoint_workspace = '/v1/workspaces/{workspace_id}'

    def __repr__(self):
        return self.get__repr__(OrderedDict({"id": self._id, "name": self._name}))

    @property
    def id(self):
        """
        :type: string
        """
        return self._id

    @property
    def name(self):
        """
        :type: string
        """
        return self._name

    def _init_properties(self):
        self._id = None
        self._name = None

    def _set_prop_values(self, data):
        if "id" in data:
            self._id = data["id"]
        if "name" in data:
            self._name = data["name"]



import requests

from retrying import retry
from .exception import AuthenticationException, AuthorisationException, InsufficientCredits, NotFound, \
    ValidationException


def if_http_exception(exception):
    return isinstance(exception, requests.exceptions.HTTPError)


class Requester:

    def __init__(self, api_key, base_url='https://api.hasty.ai', session_id=None):
        self.api_key = api_key
        self.base_url = base_url
        self.session_id = session_id
        self.headers = {
            'accept': 'application/json',
            'X-Api-Key': self.api_key,
        }
        self.cookies = dict(ui_last_state=self.session_id)

    @retry(stop_max_attempt_number=7, wait_fixed=2000, retry_on_exception=if_http_exception)
    def request(self, method, endpoint, headers, params=None, json_data=None, data=None, files=None, cookies=None):
        url = self.base_url + endpoint
        if endpoint.startswith("http"):
            url = endpoint

        response = requests.request(method,
                                    url,
                                    headers=headers,
                                    params=params,
                                    json=json_data,
                                    cookies=cookies,
                                    data=data,
                                    files=files)
        if response.status_code == 400:
            raise ValidationException(response.json().get("message"))

        if response.status_code == 401:
            raise AuthenticationException.failed_authentication()

        if response.status_code == 402:
            raise InsufficientCredits.insufficient_credits()

        if response.status_code == 403:
            raise AuthorisationException.permission_denied()

        if response.status_code == 404:
            raise NotFound.object_not_found()

        if response.status_code > 299:
            response.raise_for_status()

        return response

    def get(self, endpoint, query_params=None):
        json_data = self.request("GET", endpoint,
                                 headers=self.headers,
                                 params=query_params,
                                 cookies=self.cookies).json()
        return json_data

    def post(self, endpoint, json_data=None, content_type='application/json'):
        self.headers['Content-Type'] = content_type
        return self.request("POST", endpoint,
                            headers=self.headers,
                            json_data=json_data,
                            cookies=self.cookies).json()

    def put(self, endpoint, data=None, files=None, content_type='application/json', json_data=None):
        self.headers['Content-Type'] = content_type
        response = self.request("PUT",
                                endpoint,
                                headers=self.headers,
                                data=data,
                                files=files,
                                json_data=json_data,
                                cookies=self.cookies)

        if not data:
            if response.status_code == 200:
                return response.json()
            return None

        return None, response.status_code

    def delete(self, endpoint, json_data=None):
        response = self.request("DELETE", endpoint,
                                headers=self.headers,
                                json_data=json_data,
                                cookies=self.cookies)

        if response.status_code != 204:
            # TODO Handle different status codes
            raise Exception("Something went wrong {}", response, response.text)



from collections import OrderedDict

from .hasty_object import HastyObject


class TagClass(HastyObject):
    endpoint = '/v1/projects/{project_id}/tag_classes'
    endpoint_class = '/v1/projects/{project_id}/tag_classes/{tag_class_id}'

    def __repr__(self):
        return self.get__repr__(OrderedDict({"id": self._id, "name": self._name, "norder": self._norder}))

    @property
    def id(self):
        """
        :type: string
        """
        return self._id

    @property
    def name(self):
        """
        :type: string
        """
        return self._name

    @property
    def project_id(self):
        """
        :type: string
        """
        return self._project_id

    @property
    def norder(self):
        """
        :type: float
        """
        return self._norder

    def _init_properties(self):
        self._id = None
        self._name = None
        self._project_id = None
        self._norder = None

    def _set_prop_values(self, data):
        if "id" in data:
            self._id = data["id"]
        if "name" in data:
            self._name = data["name"]
        if "project_id" in data:
            self._project_id = data["project_id"]
        if "norder" in data:
            self._norder = data["norder"]

    @staticmethod
    def _create(requester, project_id, name, norder=None):
        res = requester.post(TagClass.endpoint.format(project_id=project_id),
                             json_data={"name": name,
                                        "norder": norder})
        return TagClass(requester, res, {"project_id": project_id})

    def edit(self, name, norder=None):
        """
        Edit tag class properties

        Arguments:
            name (str): Taп class name
            norder (float, optional): Order in the Hasty tool
        """
        self._requester.put(TagClass.endpoint_class.format(project_id=self.project_id, tag_class_id=self.id),
                            json_data={"name": name,
                                       "norder": norder})
        self._name = name
        self._norder = norder

    def delete(self):
        """
        Deletes tag class
        """
        self._requester.delete(TagClass.endpoint_class.format(project_id=self.project_id, tag_class_id=self.id))



from collections import OrderedDict
from time import time
from typing import List
import os
import logging
import urllib.request

from .constants import WAIT_INTERVAL_SEC
from .exception import ValidationException
from .hasty_object import HastyObject
from .requester import Requester


class ExportJob(HastyObject):
    endpoint = '/v1/projects/{project_id}/exports'
    endpoint_export = '/v1/projects/{project_id}/exports/{export_id}'

    def __repr__(self):
        return self.get__repr__(OrderedDict({"id": self._id, "status": self._status}))

    @property
    def id(self):
        """
        :type: string
        """
        return self._id

    @property
    def status(self):
        """
        :type: string
        """
        return self._status

    @property
    def project_id(self):
        """
        :type: string
        """
        return self._project_id

    @property
    def started_on(self):
        """
        :type: string
        """
        return self._started_on

    @property
    def completed_on(self):
        """
        :type: string
        """
        return self._completed_on

    @property
    def meta(self):
        return self._meta

    def _init_properties(self):
        self._id = None
        self._status = None
        self._project_id = None
        self._started_on = None
        self._completed_on = None
        self._meta = None
        self._last_check = None

    def _set_prop_values(self, data):
        if "id" in data:
            self._id = data["id"]
        if "status" in data:
            self._status = data["status"]
        if "started_on" in data:
            self._started_on = data["started_on"]
        if "completed_on" in data:
            self._completed_on = data["completed_on"]
        if "meta" in data:
            self._meta = data["meta"]
        if "project_id" in data:
            self._project_id = data["project_id"]

    @staticmethod
    def _create(requester: Requester, project_id: str, name: str, export_format: str,
                dataset: List[str], image_status: List[str], sign_urls: bool,
                semantic_format: str, labels_order: List[str]):
        res = requester.post(ExportJob.endpoint.format(project_id=project_id),
                             json_data={"export_name": name,
                                        "format": export_format,
                                        "dataset_id": dataset,
                                        "image_status": image_status,
                                        "sign_urls": sign_urls,
                                        "semantic_format": semantic_format,
                                        "labels_order": labels_order})
        return ExportJob(requester, res, {"project_id": project_id})

    def check_status(self):
        """
        Checks and returns the status of the export job.
        """
        if self._last_check is None or self._last_check > WAIT_INTERVAL_SEC:
            res = self._requester.get(ExportJob.endpoint_export.format(project_id=self._project_id, export_id=self._id))
            self._set_prop_values(res)
            self._last_check = time()
        return self._status

    def download(self, local_folder='.'):
        """
        Downloads export to specified folder

        Args:
             local_folder (str): Path to the local folder
        """
        self.check_status()
        if self._status == 'DONE':
            url = self._meta['url']
            filename = self._meta.get('export_name', 'export') + '.zip'
            filepath = os.path.join(local_folder, filename)
            urllib.request.urlretrieve(url, filepath)
            logging.info(f"File {filepath} saved")
        else:
            raise ValidationException.export_in_progress()



from collections import OrderedDict
from typing import List

from .exception import ValidationException
from .hasty_object import HastyObject


class Attribute(HastyObject):
    endpoint = '/v1/projects/{project_id}/attributes'
    endpoint_attribute = '/v1/projects/{project_id}/attributes/{attribute_id}'
    endpoint_attributes_classes = '/v1/projects/{project_id}/attributes_classes'

    def __repr__(self):
        return self.get__repr__(OrderedDict({"id": self._id, "name": self._name, "attribute_type": self._attribute_type,
                                             "description": self._description, "norder": self._norder,
                                             "values": self._values}))

    @property
    def id(self):
        """
        :type: string
        """
        return self._id

    @property
    def name(self):
        """
        :type: string
        """
        return self._name

    @property
    def project_id(self):
        """
        :type: string
        """
        return self._project_id

    @property
    def attribute_type(self):
        """
        :type: string
        """
        return self._attribute_type

    @property
    def description(self):
        """
        :type: string
        """
        return self._description

    @property
    def norder(self):
        """
        :type: float
        """
        return self._norder

    @property
    def values(self):
        """
        :type: Dict
        """
        return self._values

    def _init_properties(self):
        self._id = None
        self._name = None
        self._project_id = None
        self._attribute_type = None
        self._description = None
        self._norder = None
        self._values = None

    def _set_prop_values(self, data):
        if "id" in data:
            self._id = data["id"]
        if "name" in data:
            self._name = data["name"]
        if "project_id" in data:
            self._project_id = data["project_id"]
        if "type" in data:
            self._attribute_type = data["type"]
        if "description" in data:
            self._description = data["description"]
        if "norder" in data:
            self._norder = data["norder"]
        if "values" in data:
            self._values = data["values"]

    @staticmethod
    def validate_type(attribute_type):
        if attribute_type not in ['SELECTION', 'MULTIPLE-SELECTION', 'TEXT', 'INT', 'FLOAT', 'BOOL']:
            raise ValidationException('Attribute type should be one of the following '
                                      '[SELECTION, MULTIPLE-SELECTION, TEXT, INT, FLOAT, BOOL]')

    @staticmethod
    def create(requester, project_id: str, name: str, attribute_type: str, description: str, norder: float,
               values: List[str] = None):
        Attribute.validate_type(attribute_type)
        list_of_values = [{"value": v} for v in values]
        res = requester.post(Attribute.endpoint.format(project_id=project_id),
                             json_data={"name": name,
                                        "type": attribute_type,
                                        "description": description,
                                        "values": list_of_values,
                                        "norder": norder})
        return Attribute(requester, res, {"project_id": project_id})

    def edit(self, name: str, attribute_type: str, description: str, norder: float,
             values: List[str] = None):
        """
        Edit attribute properties

        Arguments:
            Args:
            name (str): Attribute name
            attribute_type (str): Attribute type ['SELECTION', 'MULTIPLE-SELECTION', 'TEXT', 'INT', 'FLOAT', 'BOOL']
            description (str, optional): Attrbute description
            norder (float, optional): Order in the Hasty tool
            values (list of str): List of values for SELECTION and MULTIPLE-SELECTION attribute type
        """
        list_of_values = [{"value": v} for v in values]
        res = self._requester.put(Attribute.endpoint_attribute.format(project_id=self.project_id, attribute_id=self.id),
                                  json_data={"name": name,
                                             "type": attribute_type,
                                             "description": description,
                                             "norder": norder,
                                             "values": list_of_values})
        self._name = res["name"]
        self._attribute_type = res["type"]
        self._description = res["description"]
        self._norder = res["norder"]
        self._values = res["values"]

    def delete(self):
        """
        Deletes attribute
        """
        self._requester.delete(Attribute.endpoint_attribute.format(project_id=self.project_id, attribute_id=self.id))

    @staticmethod
    def get_attributes_classes(requester, project_id):
        res = requester.get(Attribute.endpoint_attributes_classes.format(project_id=project_id))
        return res

    @staticmethod
    def set_attributes_classes(requester, project_id, attributes_classes):
        json_data = []
        for attr_cls in attributes_classes:
            json_data.append({"attribute_id": attr_cls["attribute_id"],
                              "class_id": attr_cls["class_id"],
                              "attribute_order": attr_cls.get("attribute_order", 0),
                              "class_order": attr_cls.get("class_order", 0)})
        res = requester.post(Attribute.endpoint_attributes_classes.format(project_id=project_id),
                             json_data=json_data)
        return res

    @staticmethod
    def delete_attributes_classes(requester, project_id, attributes_classes):
        requester.delete(Attribute.endpoint_attributes_classes.format(project_id=project_id),
                         json_data=attributes_classes)



from typing import List

from .base import Inference
from ..exception import ValidationException
from ..label_utils import check_bbox_format


class Attributer(Inference):
    model_check_endpoint = "/v1/projects/{project_id}/attributer"
    predict_endpoint = "/v1/projects/{project_id}/attribute_prediction"

    def discover_model(self):
        """
        Performs model discovery and loads model to GPU
        """
        self._discover_model(self.model_check_endpoint)

    def predict(self, image_url: str = None, image_path: str = None, bboxes: List[List[str]] = None,
                confidence_threshold: float = 0.5):
        """
        Returns predictions for provided image.

        Args:
            image_url (str): Image URL
            image_path (str): Path to local image file
            bboxes (list of list of int): List of bounding boxes [x_min, y_min, x_max, y_max]
            confidence_threshold (float): Confidence threshold [0, 1) (default 0.5)

        Returns:
            List of List of dict:
            - bbox (list of int): Coordinates of bounding box [x_min, y_min, x_max, y_max]
            - attribute_id (str): Attribute  id
            - lov_ids (list of str): List of attribute values
        """
        # Validate bboxes
        if not isinstance(bboxes, list):
            raise ValidationException("bboxes parameter must be a list of bboxes")
        for bbox in bboxes:
            if not isinstance(bbox, list):
                raise ValidationException(f"Bounding box must be a list of inegers, got {type(bbox)}")
            # Check bbox
            if bbox is not None:
                if not check_bbox_format(bbox):
                    raise ValidationException(f"Invalid bbox format - {bbox}")

        json_data = {"bboxes": bboxes,
                     "model_id": self._model_id,
                     "confidence_threshold": confidence_threshold}
        response = self._predict(image_url, image_path, Attributer.predict_endpoint,
                                 Attributer.model_check_endpoint, json_data)
        return response



from collections import OrderedDict

from hasty.hasty_object import HastyObject

from ..image import Image
from ..exception import InferenceException


class Inference(HastyObject):

    def __repr__(self):
        return self.get__repr__(OrderedDict({"model_id": self._model_id,
                                             "status": self._status}))

    @property
    def model_id(self):
        """
        :type integer
        """
        return self._model_id

    @property
    def status(self):
        """
        :type string
        """
        return self._status

    @property
    def project_id(self):
        """
        :type string
        """
        return self._project_id

    def _init_properties(self):
        self._model_id = None
        self._last_check = None
        self._status = None
        self._project_id = None

    def _set_prop_values(self, data):
        if "model_id" in data:
            self._model_id = data["model_id"]
        if "status" in data:
            self._status = data["status"]
        if "project_id" in data:
            self._project_id = data["project_id"]

    def _discover_model(self, endpoint):
        res = self._requester.get(endpoint.format(project_id=self._project_id))
        self._set_prop_values(res)

    def _predict(self, image_url: str, image_path: str, endpoint: str,
                 discover_endpoint: str, json_data):
        if self._status != 'LOADED':
            self._discover_model(discover_endpoint)
        if self._status != 'LOADED':
            raise InferenceException.model_not_loaded()
        if image_path is not None:
            url_data = Image._generate_sign_url(self._requester, self._project_id)
            with open(image_path, 'rb') as f:
                self._requester.put(url_data['url'], data=f.read(), content_type="image/*")
            upload_id = url_data['id']
            json_data['upload_id'] = upload_id
        else:
            json_data['image_url'] = image_url
        response = self._requester.post(endpoint.format(project_id=self._project_id),
                                        json_data=json_data)
        return response



class PaginatedList:

    def __init__(self, content_class, requester, endpoint, obj_params=None, query_params=None):
        self.content_class = content_class
        self.requester = requester
        self.offset = 0
        self.limit = 100
        self.query_params = query_params if query_params else dict()
        self.query_params['offset'] = self.offset
        self.query_params['limit'] = self.limit
        self.endpoint = endpoint
        self.obj_params = obj_params
        data = self.requester.get(endpoint, self.query_params)
        self.total_count = data['meta']['total']
        self.items = data['items']
        self.offset = len(self.items)
        self.instances = [content_class(self.requester, cc, obj_params) for cc in self.items]
        self.has_more = self.total_count > self.limit

    def __repr__(self):
        # TODO Add current size and total number if needed
        return str([str(i) for i in self.instances])

    def __getitem__(self, idx):
        # Check idx and fetch next batch if needed
        while len(self.items) <= idx < self.total_count:
            self.query_params['offset'] = self.offset
            new_data = self.requester.get(self.endpoint, self.query_params)
            self.items.extend(new_data["items"])
            self.instances.extend([self.content_class(self.requester, cc, self.obj_params) for cc in new_data["items"]])
            self.total_count = new_data['meta']['total']
            self.offset = len(self.items)
        return self.instances[idx]

    def __len__(self):
        return self.total_count



from hasty.attribute import Attribute
from hasty.client import Client
from hasty.dataset import Dataset
from hasty.export_job import ExportJob
from hasty.image import Image
from hasty.inference import Attributer, Detector, InstanceSegmentor, SemanticSegmentor
from hasty.label import Label
from hasty.label_class import LabelClass
from hasty.project import Project
from hasty.tag import Tag
from hasty.tag_class import TagClass
import hasty.label_utils as label_utils


def int_or_str(value):
    try:
        return int(value)
    except ValueError:
        return value


__version__ = '0.3.3'
VERSION = tuple(map(int_or_str, __version__.split('.')))

__all__ = [
    'Attribute',
    'Attributer',
    'Client',
    'Dataset',
    'Detector',
    'ExportJob',
    'Image',
    'InstanceSegmentor',
    'Label',
    'LabelClass',
    'Project',
    'SemanticSegmentor',
    'Tag',
    'TagClass',
    'label_utils'
]



from collections import OrderedDict

from .hasty_object import HastyObject
from .requester import Requester


class Dataset(HastyObject):
    endpoint = '/v1/projects/{project_id}/datasets'
    endpoint_dataset = '/v1/projects/{project_id}/datasets/{dataset_id}'

    def __repr__(self):
        return self.get__repr__(OrderedDict({"id": self._id, "name": self._name}))

    @property
    def id(self):
        """
        :type: string
        """
        return self._id

    @property
    def name(self):
        """
        :type: string
        """
        return self._name

    @property
    def project_id(self):
        """
        :type: string
        """
        return self._project_id

    @property
    def norder(self):
        """
        :type: float
        """
        return self._norder

    def _init_properties(self):
        self._id = None
        self._name = None
        self._project_id = None
        self._norder = None

    def _set_prop_values(self, data):
        if "id" in data:
            self._id = data["id"]
        if "name" in data:
            self._name = data["name"]
        if "project_id" in data:
            self._project_id = data["project_id"]
        if "norder" in data:
            self._norder = data["norder"]

    @staticmethod
    def _create(requester: Requester, project_id: str, name: str, norder: float = 0):
        res = requester.post(Dataset.endpoint.format(project_id=project_id),
                             json_data={"name": name,
                                        "norder": norder})
        return Dataset(requester, res, {"project_id": project_id})

    def edit(self, name: str, norder: float):
        """
        Edit dataset properties

        Arguments:
            name (str): Name of the dataset
            norder (float): Order in the list
        """
        res = self._requester.put(self.endpoint_dataset.format(project_id=self.project_id, dataset_id=self.id),
                                  json_data={"name": name,
                                             "norder": norder})
        self._name = res["name"]
        self._norder = res["norder"]

    def delete(self):
        """
        Removes dataset
        """
        self._requester.delete(Dataset.endpoint_dataset.format(project_id=self.project_id,
                                                               dataset_id=self.id))



from collections import OrderedDict
from typing import Dict, List, Union
import os
import urllib.request

from .constants import VALID_STATUSES
from .hasty_object import HastyObject
from .helper import PaginatedList
from .exception import ValidationException
from .label import Label
from .label_class import LabelClass
from .tag import Tag
from .tag_class import TagClass


class Image(HastyObject):
    endpoint = '/v1/projects/{project_id}/images'
    endpoint_uploads = '/v1/projects/{project_id}/image_uploads'
    endpoint_image = '/v1/projects/{project_id}/images/{image_id}'

    def __repr__(self):
        return self.get__repr__(OrderedDict({"id": self._id, "dataset_name": self._dataset_name, "name": self._name,
                                             "width": self._width, "height": self._height}))

    @property
    def id(self):
        """
        :type: string
        """
        return self._id

    @property
    def name(self):
        """
        :type: string
        """
        return self._name

    @property
    def project_id(self):
        """
        :type: string
        """
        return self._project_id

    @property
    def dataset_id(self):
        """
        :type: string
        """
        return self._dataset_id

    @property
    def dataset_name(self):
        """
        :type: string
        """
        return self._dataset_name

    @property
    def width(self):
        """
        :type: int
        """
        return self._width

    @property
    def height(self):
        """
        :type: int
        """
        return self._height

    @property
    def status(self):
        """
        :type: string
        """
        return self._status

    @property
    def public_url(self):
        """
        :type: string
        """
        return self._public_url

    def _init_properties(self):
        self._id = None
        self._name = None
        self._project_id = None
        self._dataset_name = None
        self._width = None
        self._height = None
        self._status = None
        self._public_url = None

    def _set_prop_values(self, data):
        if "id" in data:
            self._id = data["id"]
        if "name" in data:
            self._name = data["name"]
        if "project_id" in data:
            self._project_id = data["project_id"]
        if "dataset_id" in data:
            self._dataset_id = data["dataset_id"]
        if "dataset_name" in data:
            self._dataset_name = data["dataset_name"]
        if "width" in data:
            self._width = data["width"]
        if "height" in data:
            self._height = data["height"]
        if "status" in data:
            self._status = data["status"]
        if "public_url" in data:
            self._public_url = data["public_url"]

    @staticmethod
    def _get_by_id(requester, project_id, image_id):
        data = requester.get(Image.endpoint_image.format(project_id=project_id, image_id=image_id))
        return Image(requester, data, {"project_id": project_id})

    @staticmethod
    def _generate_sign_url(requester, project_id):
        data = requester.post(Image.endpoint_uploads.format(project_id=project_id), json_data={"count": 1})
        return data["items"][0]

    @staticmethod
    def _upload_from_file(requester, project_id, dataset_id, filepath):
        filename = os.path.basename(filepath)
        url_data = Image._generate_sign_url(requester, project_id)
        with open(filepath, 'rb') as f:
            requester.put(url_data['url'], data=f.read(), content_type="image/*")
        res = requester.post(Image.endpoint.format(project_id=project_id),
                             json_data={"dataset_id": dataset_id,
                                        "filename": filename,
                                        "upload_id": url_data["id"]})
        return Image(requester, res, {"project_id": project_id,
                                      "dataset_id": dataset_id})

    @staticmethod
    def _upload_from_url(requester, project_id, dataset_id, filename, url, copy_original=True):
        res = requester.post(Image.endpoint.format(project_id=project_id),
                             json_data={"dataset_id": dataset_id,
                                        "filename": filename,
                                        "url": url,
                                        "copy_original": copy_original})
        return Image(requester, res, {"project_id": project_id,
                                      "dataset_id": dataset_id})

    def get_labels(self):
        """
        Returns image labels (list of `~hasty.Label` objects)
        """
        return PaginatedList(Label, self._requester,
                             Label.endpoint_image.format(project_id=self.project_id, image_id=self.id),
                             obj_params={"project_id": self.project_id})

    def create_label(self, label_class: Union[LabelClass, str], bbox: List[int] = None, polygon: List[List[int]] = None,
                     mask: List[int] = None, z_index: int = None):
        """
        Create label

        Args:
            label_class (LabelClass, str): Label class or label class ID of the label
            bbox (list of int): Coordinates of bounding box [x_min, y_min, x_max, y_max]
            polygon (list): List of x, y pairs [[x0, y0], [x1, y1], .... [x0, y0]]
            mask (list of int): RLE Encoded binary mask, (order right -> down)
            z_index (int): Z index of the label. A label with greater value is in front of a label with a lower one.
        """
        class_id = label_class
        if isinstance(label_class, LabelClass):
            class_id = label_class.id
        label = Label._create(self._requester, self._project_id, self._id, class_id, bbox, polygon, mask, z_index)
        return label

    def create_labels(self, labels):
        """
        Create multiple labels. Returns a list of `~hasty.Label` objects

        Args:
            labels (list of dict): List of labels, keys:
                    class_id: Label class ID of the label
                    bbox: Coordinates of bounding box [x_min, y_min, x_max, y_max]
                    polygon: List of x, y pairs [[x0, y0], [x1, y1], .... [x0, y0]]
                    mask: RLE Encoded binary mask, (order right -> down)
                    z_index: Z index of the label.
        """
        return Label._batch_create(self._requester, self._project_id, self._id, labels)

    def edit_labels(self, labels):
        """
        Updates multiple labels. Returns a list of `~hasty.Label` objects

        Args:
            labels (list of dict): List of labels, keys:
                    label_id: Label id
                    class_id: Label class ID of the label
                    bbox: Coordinates of bounding box [x_min, y_min, x_max, y_max]
                    polygon: List of x, y pairs [[x0, y0], [x1, y1], .... [x0, y0]]
                    mask: RLE Encoded binary mask, (order right -> down)
                    z_index: Z index of the label.
        """
        return Label._batch_update(self._requester, self._project_id, self._id, labels)

    def delete_labels(self, label_ids: List[str]):
        """
        Removes multiple labels

        Args:
            label_ids (list of str): Returns list of ids
        """
        Label._batch_delete(self._requester, self._project_id, self._id, label_ids)

    def set_status(self, status: str):
        """
        Set image status

        Args:
            status: New status one of ["NEW", "DONE", "SKIPPED", "IN PROGRESS", "TO REVIEW"]
        """
        if status not in VALID_STATUSES:
            raise ValidationException(f"Got {status}, expected on of {VALID_STATUSES}")
        self._requester.put(Image.endpoint_image.format(project_id=self.project_id,
                                                        image_id=self.id)+"/status",
                            json_data={"status": status})

    def download(self, filepath: str):
        """
        Downloads image to file

        Args:
            filepath (str): Local path
        """
        urllib.request.urlretrieve(self._public_url, filepath)

    def get_tags(self):
        """
        Returns image tags (list of `~hasty.Tag` objects)
        """
        return PaginatedList(Tag, self._requester,
                             Tag.endpoint.format(project_id=self.project_id, image_id=self.id),
                             obj_params={"project_id": self.project_id, "image_id": self.id})

    def add_tags(self, tags: List[Union[Dict, TagClass]]):
        """
        Create multiple tags. Returns a list of `~hasty.Tag` objects

        Args:
            tags (list of dict/`~hasty.TagClass`): List of tags, keys:
                    tag_class_id: Tag class ID
        """
        return Tag._batch_create(self._requester, self._project_id, self._id, tags)

    def delete_tags(self, tags: List[Union[Dict, Tag, TagClass]]):
        """
        Removes multiple tags

        Args:
            tags (list of dict/`~hasty.Tag`/`~hasty.TagClass`): List of tags, keys:
                    tag_id: Tag ID of the label (optional if tag_class_id is specified)
                    tag_class_id: Tag class ID of the label (optional if id is specified)
        """
        Tag._batch_delete(self._requester, self._project_id, self._id, tags)



from collections import OrderedDict
from operator import itemgetter


class HastyObject:

    def __init__(self, requester, data, obj_params=None):
        self._requester = requester
        self._init_properties()
        self._set_prop_values(data)
        if obj_params:
            self._set_prop_values(obj_params)

    def get__repr__(self, params):
        """
        Converts the object to a nicely printable string.
        """

        def format_params(_params):
            items = list(_params.items())
            if not isinstance(_params, OrderedDict):
                items = sorted(items, key=itemgetter(0), reverse=True)
            for k, v in items:
                if isinstance(v, bytes):
                    v = v.decode("utf-8")
                if isinstance(v, str):
                    v = f'"{v}"'
                yield f"{k}={v}"

        return "{class_name}({params})".format(
            class_name=self.__class__.__name__,
            params=", ".join(list(format_params(params))),
        )

    def _init_properties(self):
        raise NotImplementedError()

    def _set_prop_values(self, data):
        raise NotImplementedError()



from hasty.inference.attributer import Attributer
from hasty.inference.detector import Detector
from hasty.inference.instance_segmentor import InstanceSegmentor
from hasty.inference.semantic_segmentor import SemanticSegmentor

__all__ = [
    'Attributer',
    'Detector',
    'InstanceSegmentor',
    'SemanticSegmentor'
]



from typing import Union
import uuid

from .helper import PaginatedList
from .project import Project
from .workspace import Workspace
from .requester import Requester


class Client:
    """Client for Hasty API"""

    def __init__(self, api_key, base_url='https://api.hasty.ai'):
        """
        Initialize the client

        Parameters
        ----------
        api_key
            Your API key

        """
        self.api_key = api_key
        self.base_url = base_url
        self.session_id = str(uuid.uuid4())
        self._requester = Requester(self.api_key, self.base_url, self.session_id)

    def get_workspaces(self):
        """
        Returns the list of workspaces that the user can have an access to
        :returns: A list of :py:class:`~hasty.Workspace` objects.
        """
        return PaginatedList(Workspace, self._requester, Workspace.endpoint)

    def get_projects(self):
        """
        Returns the list of projects
        :returns: A list of :py:class:`~hasty.Project` objects.
        """
        return PaginatedList(Project, self._requester, Project.endpoint)

    def get_project(self, project_id):
        """
        Returns project :py:class:`~hasty.Project` by id

        Arguments:
            project_id (str): Project id
        """
        res = self._requester.get(Project.endpoint_project.format(project_id=project_id))
        return Project(self._requester, res)

    def create_project(self, workspace: Union[str, Workspace], name: str, description: str = None) -> Project:
        """
        Creates new project :py:class:`~hasty.Project`

        Arguments:
            workspace (:py:class:`~hasty.Workspace`, str): Workspace object or id which the project should belongs to
            name (str): Name of the project
            description (str, optional): Project description
        """
        workspace_id = workspace
        if isinstance(workspace, Workspace):
            workspace_id = workspace.id
        return Project.create(self._requester, workspace_id, name, description)



from .base import Inference


class SemanticSegmentor(Inference):
    model_check_endpoint = "/v1/projects/{project_id}/semantic_segmentor"
    predict_endpoint = "/v1/projects/{project_id}/semantic_segmentation"

    def discover_model(self):
        """
        Performs model discovery and loads model to GPU
        """
        self._discover_model(self.model_check_endpoint)

    def predict(self, image_url: str = None, image_path: str = None, min_size: float = 0.0):
        """
        Returns predictions for provided image.

        Args:
            image_url (str): Image URL
            image_path (str): Path to local image file
            min_size (float): Ignore prediction with an area less then min_size

        Returns:
            List of dict:
            - bbox (list of int): Coordinates of bounding box [x_min, y_min, x_max, y_max]
            - mask (list of int): RLE Encoded binary mask, (order right -> down)
            - class_id (str): Label class ID
        """
        json_data = {"min_size": min_size,
                     "model_id": self._model_id}
        response = self._predict(image_url, image_path, SemanticSegmentor.predict_endpoint,
                                 SemanticSegmentor.model_check_endpoint, json_data)
        return response



from typing import List, Tuple

import numpy as np


# Encode mask
def rle_encoding(x: np.array):
    """
    Encode binary mask to RLE

    Args:
        x (np.array): numpy array of shape (height, width), 1 - mask, 0 - background
    Returns run length as list
    """
    dots = np.where(x.flatten() == 1)[0]  # Order right-then-down
    run_lengths = []
    prev = -2
    for b in dots:
        if b > prev + 1:
            run_lengths.extend((b + 1, 0))
        run_lengths[-1] += 1
        prev = b
    return run_lengths


# Decode mask
def rle_decode(mask_rle: List[int], shape: Tuple[int]):
    """
    Decodes mask from RLE format to binary mask. Returns numpy array, 1 - mask, 0 - background

    Args:
        mask_rle (list): run-length as string formated (start length)
        shape (tuple): (width, height) of array to return
    """
    s = mask_rle
    starts, lengths = [np.asarray(x, dtype=int) for x in (s[0:][::2], s[1:][::2])]
    starts -= 1
    ends = starts + lengths
    shape = shape[1], shape[0]
    img = np.zeros(shape[0] * shape[1], dtype=np.uint8)
    for lo, hi in zip(starts, ends):
        img[lo:hi] = 1
    return img.reshape(shape)


def yolo2hasty(yolo_bbox: List[float], image_width: int, image_height: int):
    """
    Converts normalized YOLO bbox format (Xn_center, Yn_center, Wn, Hn) to Hasty format (x_min, y_min, x_max, y_max)

    Args:
         yolo_bbox (list of floats): Normalized coordinates [X_center, Y_center, W, H] example [0.3, 0.4, 0.1, 0.25]
         image_width (int): Image width
         image_height (int): Image height
    """
    bbox = [
        int((float(yolo_bbox[1]) - float(yolo_bbox[3]) / 2) * image_width),
        int((float(yolo_bbox[2]) - float(yolo_bbox[4]) / 2) * image_height),
        int((float(yolo_bbox[1]) + float(yolo_bbox[3]) / 2) * image_width),
        int((float(yolo_bbox[2]) + float(yolo_bbox[4]) / 2) * image_height),
    ]
    return bbox


def check_bbox_format(bbox):
    if len(bbox) != 4:
        return False
    if bbox[2] <= bbox[0]:
        return False
    if bbox[3] <= bbox[1]:
        return False
    return True


def check_rle_mask(bbox, mask):
    if bbox is None:
        return False
    if mask is None:
        return False
    bbox_area = (bbox[2] - bbox[0]) * (bbox[3] - bbox[1])
    rle_area = mask[-2] - 1 + mask[-1]
    if rle_area > bbox_area:
        return False
    return True


def polygon_area(polygon):
    np_poly = np.array(polygon)
    x, y = np_poly[:, 0], np_poly[:, 1]
    return 0.5*np.abs(np.dot(x, np.roll(y, 1))-np.dot(y, np.roll(x, 1)))


def polygon2box(polygon: List[List[int]]):
    """
    Converts polygon to bounding box [x_min, y_min, x_max, y_max]

    Args:
        polygon: List of x, y pairs [[x0, y0], [x1, y1], .... [x0, y0]]
    """
    min_x, min_y, max_x, max_y = [None] * 4

    for item in polygon:
        x, y = item[0], item[1]
        if min_x is None or x < min_x:
            min_x = x
        if min_y is None or y < min_y:
            min_y = y
        if max_x is None or x > max_x:
            max_x = x
        if max_y is None or y > max_y:
            max_y = y

    bbox = [min_x, min_y, max_x, max_y]
    assert all([b is not None for b in bbox]), "Empty sequence"
    return bbox



from collections import OrderedDict
from typing import List, Union

from .attribute import Attribute
from .constants import ImageStatus, SemanticFormat, VALID_EXPORT_FORMATS, VALID_SEMANTIC_ORDER, \
    VALID_SEMANTIC_FORMATS, VALID_STATUSES
from .dataset import Dataset
from .export_job import ExportJob
from .exception import ValidationException
from .hasty_object import HastyObject
from .helper import PaginatedList
from .inference import Attributer, Detector, InstanceSegmentor, SemanticSegmentor
from .image import Image
from .label_class import LabelClass
from .tag_class import TagClass


class Project(HastyObject):
    endpoint = '/v1/projects'
    endpoint_project = '/v1/projects/{project_id}'

    def __repr__(self):
        return self.get__repr__(OrderedDict({"id": self._id, "name": self._name}))

    @property
    def id(self):
        """
        :type: string
        """
        return self._id

    @property
    def name(self):
        """
        :type: string
        """
        return self._name

    @property
    def description(self):
        """
        :type: string
        """
        return self._description

    @property
    def workspace_id(self):
        """
        :type: string
        """
        return self._workspace_id

    def _init_properties(self):
        self._id = None
        self._name = None
        self._workspace_id = None
        self._description = None

    def _set_prop_values(self, data):
        if "id" in data:
            self._id = data["id"]
        if "name" in data:
            self._name = data["name"]
        if "workspace_id" in data:
            self._workspace_id = data["workspace_id"]
        if "description" in data:
            self._description = data["description"]

    @staticmethod
    def create(requester, workspace_id, name, description):
        res = requester.post(Project.endpoint,
                             json_data={"workspace_id": workspace_id,
                                        "name": name,
                                        "description": description})
        return Project(requester, res, {"workspace_id": workspace_id})

    def edit(self, name, description):
        """
        Edits projects properties

        Arguments:
            name (str): Name of the project
            description (str, optional): Project description
        """
        res = self._requester.put(Project.endpoint_project.format(project_id=self.id),
                                  json_data={"name": name,
                                             "description": description})
        self._name = res["name"]
        self._description = res["description"]

    def delete(self):
        """
        Removes project
        """
        self._requester.delete(Project.endpoint_project.format(project_id=self.id))

    def get_datasets(self):
        """
        Returns projects datasets :py:class:`~hasty.Dataset` objects.
        """
        return PaginatedList(Dataset, self._requester,
                             Dataset.endpoint.format(project_id=self._id),
                             {"project_id": self._id})

    def get_dataset(self, dataset_id: str):
        """
        Get dataset by id, returns `~hasty.Dataset` object

        Arguments:
            dataset_id (str): Dataset id
        """
        res = self._requester.get(Dataset.endpoint_dataset.format(project_id=self.id, dataset_id=dataset_id))
        return Dataset(self._requester, res, {"project_id": self.id})

    def create_dataset(self, name: str, norder: float = 0):
        """
        Creates a new dataset, returns `~hasty.Dataset` object

        Arguments:
            name (str): Name of the dataset
            norder (float, optional): Order in the list
        """
        return Dataset._create(self._requester, self._id, name, norder)

    def get_images(self, dataset=None, image_status=None):
        """
        Retrieves the list of projects images.

        Args:
            dataset (str, `~hasty.Dataset`, list of str, list of `~hasty.Dataset`): filter images by dataset
            image_status (str, list of str): Filters images by status, valid values are:

             - "NEW"
             - "DONE",
             - "SKIPPED"
             - "IN PROGRESS"
             - "TO REVIEW"
             - "AUTO-LABELLED"

        """
        query_params = {}
        if dataset:
            if isinstance(dataset, str):
                query_params["dataset_id"] = dataset
            elif isinstance(dataset, Dataset):
                query_params["dataset_id"] = dataset.id
            elif isinstance(dataset, list):
                dataset_ids = []
                for d in dataset:
                    if isinstance(d, str):
                        dataset_ids.append(d)
                    elif isinstance(d, Dataset):
                        dataset_ids.append(d.id)
                query_params["dataset_id"] = ','.join(dataset_ids)
        if image_status:
            if isinstance(image_status, str):
                query_params["image_status"] = image_status
            elif isinstance(image_status, list):
                image_statuses = []
                for status in image_status:
                    if status in VALID_STATUSES:
                        image_statuses.append(status)
                query_params["image_status"] = ','.join(image_statuses)
        return PaginatedList(Image, self._requester,
                             Image.endpoint.format(project_id=self._id), obj_params={"project_id": self.id},
                             query_params=query_params)

    def get_image(self, image_id):
        """
        Retrieves the image by its id.

        Args:
            image_id (str): Image ID
        """
        return Image._get_by_id(self._requester, self._id, image_id)

    def upload_from_file(self, dataset, filepath):
        """
        Uploads image from the given filepath

        Args:
            dataset (`~hasty.Dataset`, str): Dataset object or id that the image should belongs to
            filepath (str): Local path
        """
        dataset_id = dataset
        if isinstance(dataset, Dataset):
            dataset_id = dataset.id
        return Image._upload_from_file(self._requester, self._id, dataset_id, filepath)

    def upload_from_url(self, dataset: Union[Dataset, str], filename: str, url: str, copy_original: bool = True):
        """
        Uploads image from a given URL

        Args:
            dataset (`~hasty.Dataset`, str): Dataset object or id that the image should belongs to
            filename (str): Filename of the image
            url (str): Image url
            copy_original (str): If True Hasty makes a copy of the image. Default True.
        """
        dataset_id = dataset
        if isinstance(dataset, Dataset):
            dataset_id = dataset.id
        return Image._upload_from_url(self._requester, self._id, dataset_id, filename, url, copy_original=copy_original)

    def get_label_classes(self):
        """
        Get label classes, list of :py:class:`~hasty.LabelClass` objects.
        """
        return PaginatedList(LabelClass, self._requester,
                             LabelClass.endpoint.format(project_id=self._id),
                             obj_params={"project_id": self.id})

    def get_label_class(self, label_class_id: str):
        """
        Get label class by id, returns `~hasty.LabelClass` object

        Arguments:
            label_class_id (str): Label class id
        """
        res = self._requester.get(LabelClass.endpoint_class.format(project_id=self.id, label_class_id=label_class_id))
        return LabelClass(self._requester, res, {"project_id": self.id})

    def create_label_class(self, name: str, color: str = None, class_type: str = "object", norder: float = None):
        """
        Create label class, returns :py:class:`~hasty.LabelClass` object.

        Args:
            name (str): Label class name
            color (str, optional): Color in HEX format #0f0f0faa
            class_type (str, optional): Class type [object or background] (default object)
            norder (float, optional): Order in the Hasty tool
        """
        return LabelClass._create(self._requester, self._id, name, color, class_type, norder)

    def get_tag_classes(self):
        """
        Get tag classes, list of :py:class:`~hasty.TagClass` objects.
        """
        return PaginatedList(TagClass, self._requester,
                             TagClass.endpoint.format(project_id=self._id),
                             obj_params={"project_id": self.id})

    def get_tag_class(self, tag_class_id: str):
        """
        Get tag class by id, returns `~hasty.TagClass` object

        Arguments:
            tag_class_id (str): Tag class id
        """
        res = self._requester.get(TagClass.endpoint_class.format(project_id=self.id, tag_class_id=tag_class_id))
        return TagClass(self._requester, res, {"project_id": self.id})

    def create_tag_class(self, name: str, norder: float = None):
        """
        Create tag class, returns :py:class:`~hasty.TagClass` object.

        Args:
            name (str): Tag class name
            norder (float, optional): Order in the Hasty tool
        """
        return TagClass._create(self._requester, self._id, name, norder)

    def get_attributes(self):
        """
        Get label classes, list of :py:class:`~hasty.Attribute` objects.
        """
        return PaginatedList(Attribute, self._requester,
                             Attribute.endpoint.format(project_id=self._id))

    def create_attribute(self, name: str, attribute_type: str, description: str, norder: float,
                         values: List[str] = None):
        """
        Create attribute, returns :py:class:`~hasty.Attribute` object.

        Args:
            name (str): Attribute name
            attribute_type (str): Attribute type ['SELECTION', 'MULTIPLE-SELECTION', 'TEXT', 'INT', 'FLOAT', 'BOOL']
            description (str, optional): Attrbute description
            norder (float, optional): Order in the Hasty tool
            values (list of str): List of values for SELECTION and MULTIPLE-SELECTION attribute type
        """
        return Attribute.create(self._requester, self._id, name, attribute_type, description, norder, values)

    def get_attribute_classes(self):
        """
        Get attributes - class mapping. Returns list of dict with a keys:
            - attribute_id - Attribute ID
            - class_id - Class ID
            - attribute_order - Order of attributes within the class
            - class_order - Order of classes within the attribute
        """
        return Attribute.get_attributes_classes(self._requester, self._id)

    def set_attribute_classes(self, attribute_classes):
        """
        Set attribute - class mapping

        Args:
            attribute_classes (dict) - list of dict with a keys:
            - attribute_id - Attribute ID
            - class_id - Class ID
            - attribute_order - Order of attributes within the class
            - class_order - Order of classes within the attribute
        """
        return Attribute.set_attributes_classes(self._requester, self._id, attribute_classes)

    def delete_attribute_classes(self, attribute_classes):
        """
        Removes attribute - class mapping

        Args:
            attribute_classes (dict) - list of dict with a keys:
            - attribute_id - Attribute ID
            - class_id - Class ID
            - attribute_order - Order of attributes within the class
            - class_order - Order of classes within the attribute
        """
        return Attribute.delete_attributes_classes(self._requester, self._id, attribute_classes)

    def export(self, name: str, export_format: str, dataset: Union[Dataset, str, List[Dataset], List[str]] = None,
               image_status: Union[str, List[str]] = ImageStatus.Done, sign_urls: bool = False,
               semantic_format: str = SemanticFormat.GS_ASC, labels_order: List[str] = None):
        """
        Initiate export job. Returns :py:class:`~hasty.ExportJob` object.

        Args:
            name (str): Name of the export file
            export_format (str): Export format one of ["json_v1.1", "semantic_png", "json_coco", "images"]
            dataset (`~hasty.Dataset`, str, list of `~hasty.Dataset`, list of str): List of the datasets to export
            image_status (list of str, str): List of the image statuses to export. Default DONE
            sign_urls (bool): Whether to generate sign urls for images. Default False
            semantic_format (str): Format for semantic_png export. ["gs_desc", "gs_asc", "class_color"]
            labels_order (list of str): Draw order for semantic_png export ["z_index", "class_type", "class_order"]
        """
        if export_format not in VALID_EXPORT_FORMATS:
            raise ValidationException(f"Wrong export format {export_format}, expected one of {VALID_EXPORT_FORMATS}")
        dataset_ids = []
        if dataset:
            if isinstance(dataset, str):
                dataset_ids.append(dataset)
            elif isinstance(dataset, Dataset):
                dataset_ids.append(dataset.id)
            elif isinstance(dataset, list) or isinstance(dataset, PaginatedList):
                for d in dataset:
                    if isinstance(d, str):
                        dataset_ids.append(d)
                    elif isinstance(d, Dataset):
                        dataset_ids.append(d.id)
        image_statuses = []
        if image_status:
            if isinstance(image_status, str) and image_status in VALID_STATUSES:
                image_statuses.append(image_status)
            elif isinstance(image_status, list):
                for status in image_status:
                    if status in VALID_STATUSES:
                        image_statuses.append(status)

        if semantic_format:
            if semantic_format not in VALID_SEMANTIC_FORMATS:
                raise ValidationException(f"Wrong semantic format {semantic_format}, "
                                          f"expected one of {VALID_SEMANTIC_FORMATS}")

        if labels_order:
            if isinstance(labels_order, str):
                if labels_order in VALID_SEMANTIC_ORDER:
                    labels_order = [labels_order]
                else:
                    raise ValidationException(f"Wrong order {labels_order} expected one of {VALID_SEMANTIC_ORDER}")
            elif isinstance(labels_order, list):
                for lo in labels_order:
                    if lo not in VALID_SEMANTIC_ORDER:
                        raise ValidationException(f"Wrong order {labels_order} expected one of {VALID_SEMANTIC_ORDER}")

        return ExportJob._create(requester=self._requester,
                                 project_id=self._id,
                                 name=name,
                                 export_format=export_format,
                                 dataset=dataset_ids,
                                 image_status=image_statuses,
                                 sign_urls=sign_urls,
                                 semantic_format=semantic_format,
                                 labels_order=labels_order)

    def get_detector(self):
        """
        Returns object detection model. Returns :py:class:`~hasty.Detector` object.
        """
        d = Detector(self._requester, {"project_id": self._id})
        d.discover_model()
        return d

    def get_instance_segmentor(self):
        """
        Returns instance segmentation model. Returns :py:class:`~hasty.InstanceSegmentor` object.
        """
        d = InstanceSegmentor(self._requester, {"project_id": self._id})
        d.discover_model()
        return d

    def get_semantic_segmentor(self):
        """
        Returns semantic segmentation model. Returns :py:class:`~hasty.SemanticSegmentor` object.
        """
        d = SemanticSegmentor(self._requester, {"project_id": self._id})
        d.discover_model()
        return d

    def get_attributer(self):
        """
        Returns active attributer model for the project. Returns :py:class:`~hasty.Attributer` object.
        """
        d = Attributer(self._requester, {"project_id": self._id})
        d.discover_model()
        return d



class ValidationException(Exception):
    def __init__(self, message):
        self.message = message

    @classmethod
    def export_in_progress(cls):
        raise ValidationException("Export is still running")


class AuthenticationException(Exception):
    def __init__(self, message):
        self.message = message

    @classmethod
    def failed_authentication(cls):
        raise AuthenticationException("Authentication failed, check your API key")


class AuthorisationException(Exception):
    def __init__(self, message):
        self.message = message

    @classmethod
    def permission_denied(cls):
        raise AuthorisationException("Looks like service account doesn't have a permission to perform this operation")


class InsufficientCredits(Exception):
    def __init__(self, message):
        self.message = message

    @classmethod
    def insufficient_credits(cls):
        raise InsufficientCredits("Looks like you out of credits, please top up your account or contact Hasty")


class NotFound(Exception):
    def __init__(self, message):
        self.message = message

    @classmethod
    def object_not_found(cls):
        raise NotFound("Referred object not found, please check your script")


class LimitExceededException(Exception):
    def __init__(self, message):
        self.message = message

    @classmethod
    def max_labels_per_batch(cls, got):
        return LimitExceededException(f"Max number of labels per batch is 100, got {got}")

    @classmethod
    def max_tags_per_batch(cls, got):
        return LimitExceededException(f"Max number of tags per batch is 100, got {got}")


class InferenceException(Exception):
    def __init__(self, message):
        self.message = message

    @classmethod
    def model_not_loaded(cls):
        return InferenceException("Model not loaded")



from .base import Inference


class InstanceSegmentor(Inference):
    model_check_endpoint = "/v1/projects/{project_id}/instance_segmentor"
    predict_endpoint = "/v1/projects/{project_id}/instance_segmentation"

    def discover_model(self):
        """
        Performs model discovery and loads model to GPU
        """
        self._discover_model(self.model_check_endpoint)

    def predict(self, image_url: str = None, image_path: str = None, confidence_threshold: float = 0.5,
                max_detections_per_image: int = 100):
        """
        Returns predictions for provided image.

        Args:
            image_url (str): Image URL
            image_path (str): Path to local image file
            confidence_threshold (float): Confidence threshold [0, 1) (default 0.5)
            max_detections_per_image (int): Maximum detections per image (default 100)

        Returns:
            List of dict:
            - bbox (list of int): Coordinates of bounding box [x_min, y_min, x_max, y_max]
            - mask (list of int): RLE Encoded binary mask, (order right -> down)
            - score (float): Confidence score
            - class_id (str): Label class ID
        """
        json_data = {"confidence_threshold": confidence_threshold,
                     "max_detections_per_image": max_detections_per_image,
                     "model_id": self._model_id}
        response = self._predict(image_url, image_path, InstanceSegmentor.predict_endpoint,
                                 InstanceSegmentor.model_check_endpoint, json_data)
        return response



from collections import OrderedDict
from .hasty_object import HastyObject


class LabelAttribute(HastyObject):
    endpoint = '/v1/projects/{project_id}/labels/{label_id}/attributes'
    set_endpoint = '/v1/projects/{project_id}/labels/{label_id}/attributes/{attribute_id}'

    def __repr__(self):
        return self.get__repr__(OrderedDict({"id": self._id, "label_id": self._label_id,
                                             "attribute_id": self._attribute_id,
                                             "value": self._value}))

    @property
    def id(self):
        """
        :type: string
        """
        return self._id

    @property
    def label_id(self):
        """
        :type: string
        """
        return self._label_id

    @property
    def project_id(self):
        """
        :type: string
        """
        return self._project_id

    @property
    def attribute_id(self):
        """
        :type: string
        """
        return self._attribute_id

    @property
    def value(self):
        """
        :type: str, int, float, bool, List
        """
        return self._value

    def _init_properties(self):
        self._id = None
        self._label_id = None
        self._project_id = None
        self._attribute_id = None
        self._value = None

    def _set_prop_values(self, data):
        if "id" in data:
            self._id = data["id"]
        if "label_id" in data:
            self._name = data["label_id"]
        if "project_id" in data:
            self._project_id = data["project_id"]
        if "attribute_id" in data:
            self._attribute_id = data["attribute_id"]
        if "value" in data:
            self._value = data["value"]

    @staticmethod
    def _set_label_attribute(requester, project_id, label_id, attribute_id, value):
        res = requester.post(LabelAttribute.set_endpoint.format(project_id=project_id, label_id=label_id,
                                                                attribute_id=attribute_id),
                             json_data={"value": value})
        return LabelAttribute(requester, res, {"project_id": project_id, "label_id": label_id,
                                               "attribute_id": attribute_id})



import codecs
import re
from setuptools import setup

with codecs.open("hasty/__init__.py") as file:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        file.read(),
        re.MULTILINE,
    ).group(1)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='hasty',
    version=version,
    description='Hasty API client library',
    url='https://github.com/hasty-ai/hasty-python/',
    author='Hasty GmbH',
    author_email="herbert@hasty.ai",
    license='MIT License',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=['hasty'],
    package_data={
        'hasty': ['inference/*']
    },
    setup_requires=[
        "numpy>=1.16",
    ],
    install_requires=["numpy>=1.16", 'requests >= 2.23.0', 'retrying==1.3.3'],
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from collections import OrderedDict
from uuid import UUID
import numbers

from .attribute import Attribute
from .hasty_object import HastyObject
from .helper import PaginatedList
from .exception import LimitExceededException, ValidationException
from .label_attribute import LabelAttribute
from .label_class import LabelClass
from .label_utils import check_bbox_format, check_rle_mask


C_LABELS_LIMIT = 100


class Label(HastyObject):
    endpoint_image = '/v1/projects/{project_id}/images/{image_id}/labels'
    endpoint_project = '/v1/projects/{project_id}/labels'

    def __repr__(self):
        return self.get__repr__(OrderedDict({"id": self._id, "bbox": self._bbox, "polygon": self._polygon}))

    def __iter__(self):
        yield 'id', self._id
        yield 'project_id', self._project_id
        yield 'image_id', self._image_id
        yield 'class_id', self._class_id
        yield 'bbox', self._bbox
        yield 'polygon', self._polygon
        yield 'mask', self._mask
        yield 'z_index', self._z_index

    @property
    def id(self):
        """
        :type: string
        """
        return self._id

    @property
    def project_id(self):
        """
        :type: string
        """
        return self._project_id

    @property
    def image_id(self):
        """
        :type: string
        """
        return self._image_id

    @property
    def class_id(self):
        """
        :type: string
        """
        return self._class_id

    @property
    def bbox(self):
        """
        :type: string
        """
        return self._bbox

    @property
    def polygon(self):
        """
        :type: string
        """
        return self._polygon

    @property
    def mask(self):
        """
        :type: string
        """
        return self._mask

    @property
    def z_index(self):
        """
        :type: string
        """
        return self._z_index

    def _init_properties(self):
        self._id = None
        self._project_id = None
        self._image_id = None
        self._class_id = None
        self._bbox = None
        self._polygon = None
        self._mask = None
        self._z_index = None

    def _set_prop_values(self, data):
        if "id" in data:
            self._id = data["id"]
        if "project_id" in data:
            self._project_id = data["project_id"]
        if "image_id" in data:
            self._image_id = data["image_id"]
        if "class_id" in data:
            self._class_id = data["class_id"]
        if "bbox" in data:
            self._bbox = data["bbox"]
        if "polygon" in data:
            self._polygon = data["polygon"]
        if "mask" in data:
            self._mask = data["mask"]
        if "z_index" in data:
            self._z_index = data["z_index"]

    @staticmethod
    def _validate_label(class_id, bbox=None, polygon=None, mask=None, z_index=None):
        try:
            class_id = UUID(class_id)
        except Exception:
            raise ValidationException(f"Invalid UUID - {class_id}")
        # Check bbox
        if bbox is not None:
            if not check_bbox_format(bbox):
                raise ValidationException(f"Invalid bbox format - {bbox}")
        # Check polygon
        if polygon is not None:
            if not isinstance(polygon, list):
                raise ValidationException(f"Polygon must be a list of [x, y] pairs - {polygon}")
            for v in polygon:
                if not isinstance(v, list) and not isinstance(v, tuple):
                    raise ValidationException(f"Polygon vertex must be a list of [x, y] - {polygon}")
                if len(v) != 2:
                    raise ValidationException(f"Polygon vertex must be a list of [x, y] - {polygon}")
        else:
            if bbox is None:
                raise ValidationException("Polygon or bounding box must be provided")
        # Check mask
        if mask is not None:
            if not check_rle_mask(bbox, mask):
                raise ValidationException(f"Invalid mask - ({bbox}, {mask})")

        if z_index:
            if not isinstance(z_index, numbers.Number):
                raise ValidationException(f"Z Index must be None or numeric, got - {z_index}")

    @staticmethod
    def _create(requester, project_id, image_id, class_id, bbox=None, polygon=None, mask=None, z_index=None):
        new_labels = Label._batch_create(requester, project_id, image_id,
                                         [{"class_id": class_id,
                                           "bbox": bbox,
                                           "polygon": polygon,
                                           "mask": mask,
                                           "z_index": z_index}])
        return new_labels[0]

    @staticmethod
    def _batch_create(requester, project_id, image_id, labels):
        if len(labels) == 0:
            return []
        data = []
        if len(labels) > C_LABELS_LIMIT:
            raise LimitExceededException.max_labels_per_batch(len(labels))
        for label in labels:
            Label._validate_label(label["class_id"], label.get("bbox"), label.get("polygon"),
                                  label.get("mask"), label.get("z_index"))
            data.append({"class_id": label["class_id"],
                         "bbox": label.get("bbox"),
                         "polygon": label.get("polygon"),
                         "mask": label.get("mask"),
                         "z_index": label.get("z_index")})
        res = requester.post(Label.endpoint_image.format(project_id=project_id, image_id=image_id), json_data=data)
        new_labels = []
        for label in res["items"]:
            new_labels.append(Label(requester, label, {"project_id": project_id, "image_id": image_id}))
        return new_labels

    @staticmethod
    def _batch_update(requester, project_id, image_id, labels):
        data = []
        if len(labels) > 100:
            raise LimitExceededException.max_labels_per_batch(len(labels))
        for label in labels:
            Label._validate_label(label["class_id"], label.get("bbox"), label.get("polygon"),
                                  label.get("mask"), label.get("z_index"))
            data.append({"label_id": label["label_id"],
                         "class_id": label["class_id"],
                         "bbox": label.get("bbox"),
                         "polygon": label.get("polygon"),
                         "mask": label.get("mask"),
                         "z_index": label.get("z_index")})
        res = requester.put(Label.endpoint_image.format(project_id=project_id, image_id=image_id), json_data=data)
        updated_labels = []
        for label in res["items"]:
            updated_labels.append(Label(requester, label, {"project_id": project_id, "image_id": image_id}))
        return updated_labels

    @staticmethod
    def _batch_delete(requester, project_id, image_id, label_ids):
        data = []
        for label_id in label_ids:
            data.append({"id": label_id})
        requester.delete(Label.endpoint_image.format(project_id=project_id, image_id=image_id), json_data=data)

    def edit(self, label_class, bbox=None, polygon=None, mask=None, z_index=None):
        """
        Update label properties

        Args:
            label_class (LabelClass, str): Label class or label class ID of the label
            bbox (list of int): Coordinates of bounding box [x_min, y_min, x_max, y_max]
            polygon (list): List of x, y pairs [[x0, y0], [x1, y1], .... [x0, y0]]
            mask (list of int): RLE Encoded binary mask, (order right -> down)
            z_index (float): Z index of the label. A label with greater value is in front of a label with a lower one.
        """
        class_id = label_class
        if isinstance(label_class, LabelClass):
            class_id = label_class.id
        updated_labels = Label._batch_update(self._requester, self.project_id, self.image_id,
                                             [{"label_id": self.id,
                                               "class_id": class_id,
                                               "bbox": bbox,
                                               "polygon": polygon,
                                               "mask": mask,
                                               "z_index": z_index}])
        updated_label = updated_labels[0]
        self._class_id = updated_label.class_id
        self._bbox = updated_label.bbox
        self._polygon = updated_label.polygon
        self._mask = updated_label.mask
        self._z_index = updated_label.z_index

    def delete(self):
        """
        Delete label
        """
        Label._batch_delete(self._requester, self.project_id, self.image_id, [self.id])

    def get_attributes(self):
        """
        Get attributes values, list of :py:class:`~hasty.LabelAttribute` objects.
        """
        return PaginatedList(LabelAttribute, self._requester,
                             LabelAttribute.endpoint.format(project_id=self.project_id, label_id=self.id))

    def set_attribute(self, attribute, value):
        """
        Set label attribute

        Args:
            attribute (`~hasty.Attribute`, str) - `~hasty.Attribute` object or attribute id
            value (str, float, int, bool, list of str) - Attribute value
        """
        attribute_id = attribute
        if isinstance(attribute, Attribute):
            attribute_id = attribute.id
        LabelAttribute._set_label_attribute(self._requester, self.project_id, self.id, attribute_id, value)
